/*
    qTUIO - TUIO Interface for Qt

    Original Version by Martin Blankenburg <martin.blankenburg@imis.uni-luebeck.de>
    Integrated into qTUIO by x29a <0.x29a.0@gmail.com>
    Some modifications by Paolo Olivo <olivopaolo@tiscali.it>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "qtuio.h"

#include <QEvent>
#include <QTouchEvent>
#include <QtGui>
#include <QtWidgets>
/*#include <QApplication>
#include <QDesktopWidget>
#include <QWidget>
#include <QGraphicsItem>
#include <QGraphicsView>
#include <QGraphicsWidget>
#include <QMainWindow>*/

#include <QtDebug>

#include "TuioClient.h"
#include "TuioCursor.h"
#include "SpeedListener.h"
#include "TouchDispatcher.h"

QTuio::QTuio(QObject *parent) : id(0) {
    theWidget = qobject_cast<QWidget *>(parent);
    theView = qobject_cast<QGraphicsView *>(parent);
    if(theView)
        theScene = theView->scene();
    else
        theScene = qobject_cast<QGraphicsScene *>(parent);
}

QTuio::~QTuio() {
    if(running) {
        tuioClient->disconnect();
        delete tuioClient;
        running = false;
        wait();
    }
}

void QTuio::run() {
    running = true;
    tuioClient = new TUIO::TuioClient(3333);
    tuioClient->addTuioListener(this);
    tuioClient->connect();
}

void QTuio::setTuioCalibration(SpeedListener *widget, QMatrix3x3 &calibration) {
    this->calibration.insert(namesSource.key(widget), calibration);
}

void QTuio::associateSourceWidget(QString name, SpeedListener *widget){
    //qDebug() << "associateSourceWidget" << name;
    namesSource.insert(name, widget);

    // Default calibration has screen size
    QRect screen = QApplication::desktop()->rect();
    float data[] = { screen.width(), 0, QApplication::desktop()->x(),
                     0, screen.height(), QApplication::desktop()->y(),
                     0, 0, 1.0 } ;
    calibration.insert(name, QMatrix3x3(data));
}

void QTuio::dissociateWidget(QString name){
    qDebug() << "dissociateWidget" << name;
    namesSource.remove(name);
    calibration.remove(name);
}

int total = 0;

void QTuio::addTuioCursor(TUIO::TuioCursor *tcur) {
    total ++;
    QString name = tcur->getTuioSourceName();
    name += tcur->getTuioSourceAddress();
    if(!namesSource.contains(name))
        newSource(name, tcur->getTuioSourceAddress());
    QWidget *target ;
    QPoint screenPos = tuioToScreen(name, QPointF(tcur->getX(), tcur->getY())) ;
    target = theWidget->childAt((int)screenPos.x() - theWidget->geometry().x(),
                                (int)screenPos.y() - theWidget->geometry().y()) ;
    if(target == 0) target = theWidget ;

    QTouchEvent::TouchPoint touchPoint(id++) ;
    touchPoint.setRect(QRectF()) ;
    touchPoint.setPressure(1.0) ;
    touchPoint.setState(Qt::TouchPointPressed) ;

    struct Tuio2Qt qts = {touchPoint, target} ;
    QString deviceId = tcur->getTuioSourceAddress();
    deviceId += tcur->getSessionID();
    tuio2Qt.insert(deviceId, qts) ;
    // Associate the new touchPoint to the correspondent widget
    QList<QString> widgetTouches;
    if(widgets.contains(target))
        widgetTouches = widgets[target] ;
    widgetTouches.append(deviceId) ;
    widgets.insert(target, widgetTouches) ;
    updateTouch(tcur) ;
}

void QTuio::updateTuioCursor(TUIO::TuioCursor *tcur) {
    QString deviceId = tcur->getTuioSourceAddress();
    deviceId += tcur->getSessionID();
    if(!tuio2Qt.contains(deviceId))
        return;
    struct Tuio2Qt qts = tuio2Qt[deviceId] ;
    QTouchEvent::TouchPoint &tPoint = qts.tPoint ;
    if(tPoint.state() != Qt::TouchPointPressed){
        //if(tcur->getMotionSpeed() > 0){
            /*foreach(SpeedListener *sl, namesSource.values())
                sl->setSpeedForTouch(tPoint.id(), QPointF(tcur->getXSpeed(), tcur->getYSpeed()));*/
            QString name = tcur->getTuioSourceName();
            name += tcur->getTuioSourceAddress();
            SpeedListener *sl = namesSource.value(name);
            if(sl)
                sl->setSpeedForTouch(tPoint.id(), QPointF(tcur->getXSpeed(), tcur->getYSpeed()));

            tPoint.setState(Qt::TouchPointMoved);
        /*}else{
            tPoint.setState(Qt::TouchPointStationary);
        }*/
    }
    tuio2Qt.insert(deviceId, qts) ;
    updateTouch(tcur) ;
}

void QTuio::removeTuioCursor(TUIO::TuioCursor *tcur) {
    total--;
    QString deviceId = tcur->getTuioSourceAddress();
    deviceId += tcur->getSessionID();
    if(!tuio2Qt.contains(deviceId))
        return;
    struct Tuio2Qt qts = tuio2Qt[deviceId] ;
    qts.tPoint.setState(Qt::TouchPointReleased) ;
    updateTouch(tcur) ;
    tuio2Qt.insert(deviceId, qts) ;
    /*foreach(SpeedListener *sl, namesSource.values())
        sl->removeSpeedTouch(qts.tPoint.id());*/
    QString name = tcur->getTuioSourceName();
    name += tcur->getTuioSourceAddress();
    SpeedListener *sl = namesSource.value(name);
    if(sl)
        sl->removeSpeedTouch(qts.tPoint.id());

}

void QTuio::addTuioObject(TUIO::TuioObject * /*tobj*/) {}
void QTuio::updateTuioObject(TUIO::TuioObject * /*tobj*/) {}
void QTuio::removeTuioObject(TUIO::TuioObject * /*tobj*/) {}

void QTuio::addTuioBlob(TUIO::TuioBlob */*tblb*/) {}
void QTuio::updateTuioBlob(TUIO::TuioBlob */*tblb*/) {}
void QTuio::removeTuioBlob(TUIO::TuioBlob */*tblb*/) {}

void  QTuio::refresh(TUIO::TuioTime /*frameTime*/) {
    //  qDebug("refresh") ;
    foreach(QWidget *currWidget, widgets.keys()){
        bool touchBegin = true ;
        bool touchEnd = true ;
        bool sendEvent = false ;
        Qt::TouchPointStates tps ;
        QList<QTouchEvent::TouchPoint> tPts ;
        QList<int> listId;

        foreach(QString deviceId, widgets[currWidget]){
            struct Tuio2Qt qts = tuio2Qt[deviceId] ;
            if(qts.tPoint.state() != Qt::TouchPointPressed)
                        touchBegin = false ;
            if(qts.tPoint.state() != Qt::TouchPointReleased)
                touchEnd = false ;
            if(qts.tPoint.state() != Qt::TouchPointStationary)
                sendEvent = true ;
            tps |= qts.tPoint.state() ;
            tPts.append(qts.tPoint) ;
            listId.append(qts.tPoint.id());

            if(qts.tPoint.state() == Qt::TouchPointReleased){
                tuio2Qt.remove(deviceId) ;
                widgets[currWidget].removeAll(deviceId);
            }else{
                qts.tPoint.setState(Qt::TouchPointMoved) ;
                tuio2Qt.insert(deviceId, qts) ;
            }
        }

        /*//TODO Is it necessary ??
        //Add touches already exist in ghostview
        TouchDispatcher *ghostview = dynamic_cast<TouchDispatcher *>(theWidget);
        if(ghostview){
            foreach(QTouchEvent::TouchPoint pt, ghostview->getTouchpointList()){
                if(pt.state() != Qt::TouchPointReleased && !listId.contains(pt.id())){
                    touchBegin = false;
                    touchEnd = false;
                    pt.setState(Qt::TouchPointStationary);
                    tps |= pt.state();
                    tPts.append(pt);
                }
            }
        }*/

        if(sendEvent){
            QEvent::Type eventType ;
            if(touchBegin)
                eventType = QEvent::TouchBegin ;
            else if(touchEnd)
                eventType = QEvent::TouchEnd ;
            else
                eventType = QEvent::TouchUpdate ;

            QEvent *touchEvent = new QTouchEvent(eventType, new QTouchDevice() /*QTouchDevice::TouchScreen*/, Qt::NoModifier, tps, tPts);
            if(theView && theView->viewport())
                qApp->postEvent(theView->viewport(), touchEvent);
            else if(theScene)
                qApp->postEvent(theScene, touchEvent);
            else
                qApp->postEvent(currWidget, touchEvent);
        }
    }

    //if a widget has an empty list it can be dropped
    foreach(QWidget *currWidget, widgets.keys())
        if(widgets[currWidget].isEmpty())
            widgets.remove(currWidget);

    // if there is no more touchPoints id can be reset to zero
    if(widgets.empty()) id = 0 ;
}

void QTuio::updateTouch(TUIO::TuioCursor *tcur) {
    QString deviceId = tcur->getTuioSourceAddress();
    deviceId += tcur->getSessionID();
    if(!tuio2Qt.contains(deviceId))
        return;
    struct Tuio2Qt qts = tuio2Qt[deviceId] ;
    QTouchEvent::TouchPoint &tPoint = qts.tPoint ;

    // --- Normalized coordinates ---
    tPoint.setLastNormalizedPos(tPoint.normalizedPos()) ;
    tPoint.setNormalizedPos(QPointF(tcur->getX(), tcur->getY())) ;

    // --- Screen coordinates ---
    tPoint.setLastScreenPos(tPoint.screenPos()) ;
    QString name = tcur->getTuioSourceName();
    name += tcur->getTuioSourceAddress();
    tPoint.setScreenPos(tuioToScreen(name, tPoint.normalizedPos())) ;

    // --- Screen coordinates ---
    tPoint.setLastPos(tPoint.pos()) ;
    tPoint.setPos(mapWidgetToGlobal(qts.widget, tPoint.screenPos())) ;
    //  qDebug() << tPoint.normalizedPos() << tPoint.screenPos() << tPoint.pos() ;
    // --- Start coordinates ---
    if (tPoint.state() == Qt::TouchPointPressed) {
        tPoint.setStartNormalizedPos(tPoint.normalizedPos());
        tPoint.setStartPos(tPoint.pos());
        tPoint.setStartScreenPos(tPoint.screenPos());
        tPoint.setStartScenePos(tPoint.scenePos());

        tPoint.setLastNormalizedPos(tPoint.normalizedPos());
        tPoint.setLastPos(tPoint.pos());
        tPoint.setLastScreenPos(tPoint.screenPos());
        tPoint.setLastScenePos(tPoint.scenePos());
    }

    tuio2Qt.insert(deviceId, qts) ;
}

QPoint QTuio::tuioToScreen(QString name, const QPointF &normalized) const {
    float pos[] = { normalized.x(),
                    normalized.y(),
                    1 } ;
    QGenericMatrix<1,3,float> result;
    if(calibration.contains(name)){
        result = calibration.value(name) * QGenericMatrix<1,3,float>(pos) ;
    }else{
        // Default calibration has screen size
        QRect screen = QApplication::desktop()->rect();
        float data[] = { screen.width()/**//2.0/**/, 0, QApplication::desktop()->x(),
                         0, screen.height()/**//2.0/**/, QApplication::desktop()->y(),
                         0, 0, 1.0 } ;
        result = QMatrix3x3(data) * QGenericMatrix<1,3,float>(pos) ;
    }
    return QPoint(result.data()[0], result.data()[1]) ;
}

QPointF QTuio::mapWidgetToGlobal(QWidget *widget, const QPointF &pos) const {
    //  qDebug() << widget << pos ;
    QPointF upLevel(pos.x() - widget->geometry().x(),
                    pos.y() - widget->geometry().y());
    if (widget->parentWidget())
        return mapWidgetToGlobal(widget->parentWidget(), upLevel) ;
    else
        return upLevel ;
}
