#include "mingetframe.h"

#include <QGraphicsScene>
#include <QGraphicsSceneMouseEvent>
#include <QPainter>
#include <QStyleOption>
#include <QPropertyAnimation>
#include <math.h>

#include "abstractwidget.h"
#include "mingetclose.h"
#include "mingetsettings.h"
#include "mingeticonify.h"
#include "ghostview.h"

#define ADJUST 10

#define THICKNESS (GhostView::scaleFactor*8/GhostView::px2mmScreen)
#define BORDER_RADIUS (GhostView::scaleFactor*7/GhostView::px2mmScreen)
#define SHADOW (GhostView::scaleFactor*3/GhostView::px2mmScreen)

#define BORDER_LEFT 1
#define BORDER_RIGHT 2
#define BORDER_BOTTOM 4
#define BORDER_TOP 8


MingetFrame::MingetFrame(AbstractWidget *widget, QSize dim) : QGraphicsObject(), controlWidget(widget)
{
    setAcceptTouchEvents(true);
    setOpacity(0.6);
    x_width = dim.width()+2*THICKNESS;
    y_height = dim.height()+2*THICKNESS;
    ratio = dim.width() / (float)dim.height();
    settingsOn = false;
    iconifyOn = false;
    mode = 0;
    border = 0;

    minClose = new MingetClose(this);
    minClose->setParentItem(this);
    minSettings = new MingetSettings(this);
    minSettings->setParentItem(this);
    minIconify = new MingetIconify(this);
    minIconify->setParentItem(this);

    connect(this, SIGNAL(destroyed()), controlWidget, SLOT(deleteLater()));
    connect(this, SIGNAL(destroyed()), minClose, SLOT(deleteLater()));
    connect(this, SIGNAL(destroyed()), minSettings, SLOT(deleteLater()));
    connect(this, SIGNAL(destroyed()), minIconify, SLOT(deleteLater()));

    controlWidget->setWidth(dim.width());
    controlWidget->setHeight(dim.height());
}

void MingetFrame::toggleIconify(){
    iconifyOn = !iconifyOn;
    settingsOn = false;

    float minSize = GhostView::scaleFactor* 15.0 / GhostView::px2mmScreen;  //default minimized width = 15mm
    float scale = minSize / controlWidget->getWidth();

    if(iconifyOn){
        controlWidget->setAcceptTouchEvents(false);
        minClose->hide();
        minSettings->hide();

        QPropertyAnimation *animation = new QPropertyAnimation(this, "scale");
        animation->setDuration(1000);
        animation->setStartValue(1);
        animation->setEndValue(scale);
        animation->setEasingCurve(QEasingCurve::OutCubic);
        animation->start(QAbstractAnimation::DeleteWhenStopped);

        QPropertyAnimation *animation2 = new QPropertyAnimation(minIconify, "scale");
        animation2->setDuration(1000);
        animation2->setStartValue(1);
        animation2->setEndValue(1.0/scale);
        animation2->setEasingCurve(QEasingCurve::InOutQuad);
        animation2->start(QAbstractAnimation::DeleteWhenStopped);
    }else{
        controlWidget->setAcceptTouchEvents(true);
        minClose->show();
        minSettings->show();
        //minIconify->setScale(1);

        QPropertyAnimation *animation = new QPropertyAnimation(this, "scale");
        animation->setDuration(1000);
        animation->setStartValue(scale);
        animation->setEndValue(1);
        animation->setEasingCurve(QEasingCurve::OutCubic);
        animation->start(QAbstractAnimation::DeleteWhenStopped);

        QPropertyAnimation *animation2 = new QPropertyAnimation(minIconify, "scale");
        animation2->setDuration(1000);
        animation2->setStartValue(1.0/scale);
        animation2->setEndValue(1);
        animation2->setEasingCurve(QEasingCurve::OutExpo);
        animation2->start(QAbstractAnimation::DeleteWhenStopped);
    }
}

QRectF MingetFrame::boundingRect() const
{
    return QRectF(-x_width/2.0 - SHADOW - ADJUST, -y_height/2.0 - SHADOW - ADJUST,
                  x_width + 2*SHADOW + 2*ADJUST,    y_height + 2*SHADOW + 2*ADJUST);
}

QPainterPath MingetFrame::shape() const
{
    QPainterPath path;
    if(settingsOn)
        path.addRoundedRect(-x_width/2.0, -y_height/2.0, x_width, y_height, 27, 27);
    else
        path.addRect(-x_width/2.0+THICKNESS, -y_height/2.0+THICKNESS, x_width-2*THICKNESS, y_height-2*THICKNESS);
    return path;
}

void MingetFrame::paint(QPainter *painter, const QStyleOptionGraphicsItem *option, QWidget *)
{
    QPen pen;

    if(settingsOn){
        //Frame
        QRadialGradient gradFrame(0,0 , x_width/2.0, 0,0);
        gradFrame.setColorAt(1, QColor(150,150,150));
        gradFrame.setColorAt(0, QColor(100,100,100));
        pen.setBrush(gradFrame);
        //pen.setBrush(QColor(100, 100, 100));
        pen.setWidth(THICKNESS);
        painter->setPen(pen);
        painter->drawRoundedRect(-(x_width-THICKNESS)/2.0, -(y_height-THICKNESS)/2.0, x_width-THICKNESS, y_height-THICKNESS, BORDER_RADIUS-THICKNESS/2.0, BORDER_RADIUS-THICKNESS/2.0);
    }

    //Shadow
    QPainterPath shadow;
    if(settingsOn){
        shadow.addRoundedRect(shape().boundingRect().adjusted(-SHADOW,-SHADOW,SHADOW,SHADOW), 35,35);
    }else if(iconifyOn){
        float minShadow = 1/GhostView::px2mmScreen;
        shadow.addRect(shape().boundingRect().adjusted(-minShadow,-minShadow,minShadow,minShadow));
    }else{
        shadow.addRect(shape().boundingRect().adjusted(-SHADOW,-SHADOW,SHADOW,SHADOW));
    }
    //painter->fillPath(shadow.subtracted(shape()), QColor(63, 63, 63, 180));
    QRadialGradient gradShadow(0,0 , shadow.boundingRect().width()/2.0, 0,0);
    gradShadow.setColorAt(0, QColor(0,0,0, 200));
    gradShadow.setColorAt(1, QColor(200,200,200, 200));
    painter->fillPath(shadow.subtracted(shape()), gradShadow);

    //Shape
    pen.setBrush(QColor(255, 255, 255));
    pen.setWidth(5);
    painter->setPen(pen);
    painter->drawPath(shape());

    //Touched borders
    QPainterPath path2;
    path2.setFillRule(Qt::WindingFill);
    if((border & BORDER_RIGHT) == BORDER_RIGHT)
        path2.addEllipse(QPointF((x_width-THICKNESS)/2.0, 0), THICKNESS/2.0, (y_height-THICKNESS)/2.0);
    if((border & BORDER_LEFT) == BORDER_LEFT)
        path2.addEllipse(QPointF(-(x_width-THICKNESS)/2.0, 0), THICKNESS/2.0, (y_height-THICKNESS)/2.0);
    if((border & BORDER_BOTTOM) == BORDER_BOTTOM)
        path2.addEllipse(QPointF(0, (y_height-THICKNESS)/2.0), (x_width-THICKNESS)/2.0, THICKNESS/2.0);
    if((border & BORDER_TOP) == BORDER_TOP)
        path2.addEllipse(QPointF(0, -(y_height-THICKNESS)/2.0), (x_width-THICKNESS)/2.0, THICKNESS/2.0);
    painter->fillPath(path2, QColor(255, 255, 255));
}

bool MingetFrame::sceneEvent(QEvent *event)
{
    QTouchEvent *touchEvent;

    switch(event->type()){
    case QEvent::TouchBegin:
    case QEvent::TouchUpdate:
    case QEvent::TouchEnd:
        touchEvent = static_cast<QTouchEvent *>(event);
        foreach(QTouchEvent::TouchPoint touchPoint1, touchEvent->touchPoints()){
            switch(touchPoint1.state()){
            case Qt::TouchPointPressed:
                if(iconifyOn){
                    foreach(Cursor *cursor, controlWidget->getCursors().values())
                        cursor->setLeftClick(true);
                    break;
                }
                if(mode <= 10){
                    int border2 = 0;
                    if(touchPoint1.pos().x() >= x_width/2.0 - THICKNESS)
                        border2 = BORDER_RIGHT;
                    else if(touchPoint1.pos().x() <= -x_width/2.0 + THICKNESS)
                        border2 = BORDER_LEFT;
                    else if(touchPoint1.pos().y() >= y_height/2.0 - THICKNESS)
                        border2 = BORDER_BOTTOM;
                    else if(touchPoint1.pos().y() <= -y_height/2.0 + THICKNESS)
                        border2 = BORDER_TOP;

                    if(border != border2)
                        border += border2;

                    if(mode == 0){  //1er doigt
                        mode = 10;
                    }else if(mode == 10){  //2eme doigt
                        //1 doigt sur 2 bords opposes = redimensionnement
                        if(border == BORDER_LEFT + BORDER_RIGHT){  //horizontal
                            mode = 11;
                        }else if(border == BORDER_TOP + BORDER_BOTTOM){  //vertical
                            mode = 12;
                        }else{  //sinon rotation + deplacement
                            mode = 13;
                        }
                    }
                }
                break;

            case Qt::TouchPointMoved:
                if(iconifyOn){
                    moveBy(touchPoint1.scenePos().x() - touchPoint1.lastScenePos().x(),
                           touchPoint1.scenePos().y() - touchPoint1.lastScenePos().y());
                    break;
                }
                if(mode == 10){
                    //Move the widget
                    moveBy(touchPoint1.scenePos().x() - touchPoint1.lastScenePos().x(),
                           touchPoint1.scenePos().y() - touchPoint1.lastScenePos().y());

                //1 doigt sur 2 bords opposes = redimensionnement
                }else if(mode == 11){  //horizontal
                    x_width += abs(touchPoint1.pos().x()) - abs(touchPoint1.lastPos().x());
                    //y_height = x_width / ratio;
                    if(y_height < 4*THICKNESS){
                        y_height = 4*THICKNESS;
                        //x_width = y_height * ratio;
                    }
                    controlWidget->setWidth(x_width - 2*THICKNESS);
                    controlWidget->setHeight(y_height - 2*THICKNESS);
                    moveBy((touchPoint1.scenePos().x() - touchPoint1.lastScenePos().x())/2.0, (touchPoint1.scenePos().y() - touchPoint1.lastScenePos().y())/2.0);

                }else if(mode == 12){  //vertical
                    y_height += abs(touchPoint1.pos().y()) - abs(touchPoint1.lastPos().y());
                    //x_width = y_height * ratio;
                    if(y_height < 4*THICKNESS){
                        y_height = 4*THICKNESS;
                        //x_width = y_height * ratio;
                    }
                    controlWidget->setWidth(x_width - 2*THICKNESS);
                    controlWidget->setHeight(y_height - 2*THICKNESS);
                    moveBy((touchPoint1.scenePos().x() - touchPoint1.lastScenePos().x())/2.0, (touchPoint1.scenePos().y() - touchPoint1.lastScenePos().y())/2.0);

                }else if(mode == 13){  //sinon rotation + deplacement
                    if(touchEvent->touchPoints().size() == 2){
                        QTouchEvent::TouchPoint touchPoint2 = touchEvent->touchPoints().first();
                        if(touchPoint2.id() == touchPoint1.id())
                            touchPoint2 = touchEvent->touchPoints().at(1);
                        QLineF line1(touchPoint2.scenePos(), touchPoint1.lastScenePos());
                        QLineF line2(touchPoint2.scenePos(), touchPoint1.scenePos());

                        float dAngle = line2.angleTo(line1);
                        //QPointF dTrans = (touchPoint1.scenePos() - touchPoint1.lastScenePos())/2.0;

                        //Position de l'origine de l'image dans le repere 0
                        QPointF origine0 = scenePos();
                        //1) Determination du repere 1 associe au segment [lastA, lastB]
                        //1.1) Calcul de la matrice de passage du repere 0 au repere 1
                        QTransform m01(line1.dx(), line1.dy(),
                                       -line1.dy(), line1.dx(),
                                       touchPoint1.lastScenePos().x(), touchPoint1.lastScenePos().y());
                        //1.2) Calcul de origine0 dans repere 1
                        QTransform m10 = m01.inverted();
                        QPointF origine1 = m10.map(origine0);

                        //2) Calcul du repere 2 qui est celui associe au segment [currentA, currentB]
                        QTransform m02(line2.dx(), line2.dy(),
                                       -line2.dy(), line2.dx(),
                                       touchPoint1.scenePos().x(), touchPoint1.scenePos().y());
                        //La position de l'origine est la meme dans le repere 2 que celle dans le repere 1 au changement d'echelle pres.
                        //m02 = m02.scale(dScale, dScale);  //pris en compte dans la construction de m02
                        QPointF origine2 = origine1;
                        QPointF origine0b = m02.map(origine2);
                        QPointF dTrans = origine0b - origine0;

                        moveBy(dTrans.x(), dTrans.y());
                        setRotation((rotation() + dAngle));
                    }
                }
                widthChanged();
                break;

            case Qt::TouchPointReleased:
                if(iconifyOn){
                    foreach(Cursor *cursor, controlWidget->getCursors().values())
                        cursor->setLeftClick(false);
                    break;
                }
                if(mode > 10){
                    mode = 10;
                    int border2 = 0;
                    if(touchPoint1.pos().x() >= x_width/2.0 - THICKNESS)
                        border2 = BORDER_RIGHT;
                    else if(touchPoint1.pos().x() <= -x_width/2.0 + THICKNESS)
                        border2 = BORDER_LEFT;
                    else if(touchPoint1.pos().y() >= y_height/2.0 - THICKNESS)
                        border2 = BORDER_BOTTOM;
                    else if(touchPoint1.pos().y() <= -y_height/2.0 + THICKNESS)
                        border2 = BORDER_TOP;
                    if(border != border2)
                        border -= border2;
                }else if(mode == 10){
                    mode = 0;
                    border = 0;
                }
                controlWidget->callNeedTuioRecalibrate();
                break;

            default:
                break;
            }
        }
        break;

    default:
        return QGraphicsItem::sceneEvent(event);
    }
    return true;
}
