#include "softtouchpad.h"

#include <QGraphicsScene>
#include <QGraphicsSceneMouseEvent>
#include <QPainter>
#include <QStyleOption>
#include <QPropertyAnimation>
#include <QDebug>
#include <math.h>

#include "ghostview.h"
#include "cursor.h"
#include "link.h"
#include "mingetclose.h"
#include "mingetsettings.h"
#include "mingetframe.h"

#define BUTTON_LINE (GhostView::scaleFactor*1/GhostView::px2mmScreen)
#define BUTTON_HEIGHT GhostView::circleSize
#define MIN_DP (5*x_width/sizeTouchpad.width())


SoftTouchpad::SoftTouchpad(GhostView *g, QColor col, QPointF widgetPos, QPointF cursorPos, QSize dim) : AbstractWidget(g, col), controlMouse(false), directTouch(false), timerClutch(NULL)
{
    QLineF line(cursorPos, widgetPos);
    float angle = line.angleTo(QLineF(0,0,0,1));  //TODO fausse le déplacement

    MingetFrame *frame = new MingetFrame(this, dim);
    ghostview->scene()->addItem(frame);
    frame->setPos(widgetPos);
    frame->setRotation(angle);
    frame->setZValue(1);
    this->setParentItem(frame);
    frame->setAcceptedMouseButtons(0);
    setZValue(1.1);

    connect(frame, SIGNAL(rotationChanged()), this, SLOT(frameRotate()));

    QPropertyAnimation *animation = new QPropertyAnimation(frame, "scale");
    animation->setDuration(1000);
    animation->setStartValue(0);
    animation->setEndValue(1);
    animation->setEasingCurve(QEasingCurve::OutBack);
    animation->start(QAbstractAnimation::DeleteWhenStopped);

    //Create new cursor in center of screen
    createCursor(cursorPos);
    cursors.begin().value()->addCursel(QPoint());
    //TODO FIX bug of creating cursel before any click: crash
    cursors.begin().value()->setLeftClick(true);
    cursors.begin().value()->setLeftClick(false);
}

QPainterPath SoftTouchpad::shape() const
{
    QPainterPath path;
    path.addRect(-x_width/2.0, -y_height/2.0, x_width, y_height);
    return path;
}

void SoftTouchpad::underPaint(QPainter *painter, const QStyleOptionGraphicsItem *option, QWidget *)
{
    //Touch surface
    painter->fillRect(-x_width/2, -y_height/2, x_width, y_height-BUTTON_HEIGHT-BUTTON_LINE, QColor("white"));

    //RadarView surface
        QPainterPath path;
        path.addRect(-x_width/2, -y_height/2, x_width, y_height-BUTTON_HEIGHT-BUTTON_LINE);
        painter->setClipPath(path, Qt::ReplaceClip);
        painter->translate(0, -(BUTTON_HEIGHT+BUTTON_LINE)/2.0);  //Center view (cause of button)
        painter->rotate(-parentItem()->rotation());
        painter->drawPixmap(-ghostview->mapToGlobal(cursors.begin().value()->pos().toPoint())+ghostview->mapToGlobal(QPoint()), ghostview->getPrintScreen());
        painter->rotate(parentItem()->rotation());
        painter->translate(0, (BUTTON_HEIGHT+BUTTON_LINE)/2.0);
        painter->setClipping(false);

    //Buttons
    QPen pen;
    pen.setBrush(QColor(100, 100, 100));
    pen.setWidth(BUTTON_LINE);
    painter->setPen(pen);
    painter->drawLine(-x_width/2.0+BUTTON_LINE/2.0, y_height/2.0-BUTTON_HEIGHT-BUTTON_LINE/2.0,
                       x_width/2.0-BUTTON_LINE/2.0, y_height/2.0-BUTTON_HEIGHT-BUTTON_LINE/2.0);

    QPainterPath b1;
    b1.addRect(-x_width/2.0, y_height/2.0-BUTTON_HEIGHT,
               x_width, BUTTON_HEIGHT);

    QLinearGradient grad(0, y_height/2.0-BUTTON_HEIGHT, 0, y_height/2.0);
    if(memoryLeftClick != -1){
        grad.setFinalStop(0, y_height/2.0-BUTTON_HEIGHT/2.0);
        grad.setColorAt(1, color.darker());
        grad.setColorAt(0, color);
        //painter->fillPath(b1, QColor(0, 50, 150));
    }else{
        grad.setColorAt(1, color);
        grad.setColorAt(0, color.lighter());
        //painter->fillPath(b1, QColor(0, 100, 150));
    }
    painter->fillPath(b1, grad);
}

bool SoftTouchpad::sceneEvent(QEvent *event)
{
    QTouchEvent *touchEvent;

    switch(event->type()){
    case QEvent::TouchBegin:
    case QEvent::TouchUpdate:
    case QEvent::TouchEnd:
        touchEvent = static_cast<QTouchEvent *>(event);
        //qDebug("event(%d) received %d touch", event->type(), touchEvent->touchPoints().size());
        foreach(QTouchEvent::TouchPoint touchPoint1, touchEvent->touchPoints()){
            switch(touchPoint1.state()){

            case Qt::TouchPointPressed:
                receiveTouchBegin();
                //Buttons area
                if(touchPoint1.pos().y() > y_height/2.0-BUTTON_HEIGHT-BUTTON_LINE/2.0){
                    if(memoryLeftClick == -1){
                        memoryLeftClick = touchPoint1.id();
                        cursors.begin().value()->setLeftClick(true);
                    }

                //Control of cursor
                }else if(!controlMouse){
                    if(!directTouch){
                        //qDebug("arm cursel");
                        QTimer *timerClick = new QTimer(this);
                        timerClick->setSingleShot(true);
                        connect(timerClick, SIGNAL(timeout()), this, SLOT(startCursel()));
                        timerClickList.insert(touchPoint1.id(), timerClick);
                        clickPosition[touchPoint1.id()] = touchPoint1.pos();
                        if(timerClickList.size() == 1)
                            timerClick->start(500);
                        else{
                            //Start immediately cursel mode on second finger
                            timerClickList.begin().value()->start(0);
                            timerClick->start(0);
                        }
                    }else{
                        //Next cursel appears instantaneously
                        //qDebug("new cursel");
                        QTimer *timerClick = new QTimer(this);
                        timerClick->setSingleShot(true);
                        connect(timerClick, SIGNAL(timeout()), this, SLOT(startCursel()));
                        timerClickList.insert(touchPoint1.id(), timerClick);
                        clickPosition[touchPoint1.id()] = touchPoint1.pos();
                        timerClick->start(0);
                    }
                }
                break;

            case Qt::TouchPointMoved:
            //case Qt::TouchPointStationary:
                if(memoryLeftClick == touchPoint1.id()){

                }else{
                    //Begin move mouse
                    if(QLineF(touchPoint1.pos(), touchPoint1.startPos()).length() >= MIN_DP
                             && timerClickList.contains(touchPoint1.id())){
                        //qDebug("begin mouse");
                        timerClickList[touchPoint1.id()]->deleteLater();
                        timerClickList.remove(touchPoint1.id());
                        clickPosition.remove(touchPoint1.id());
                        QTime *time = new QTime();
                        time->start();
                        times.insert(touchPoint1.id(), time);
                        controlMouse = true;

                    //Direct touch
                    }else if(touchIdTOcurselId.contains(touchPoint1.id())){
                        cursors.begin().value()->moveCursel(touchIdTOcurselId[touchPoint1.id()], (touchPoint1.pos() - touchPoint1.lastPos())/GhostView::scaleFactor);
                    }

                    if(controlMouse){
                        float speed = 0;
                        if(speedForTouch.contains(touchPoint1.id())){
                            QPointF normalSpeed = speedForTouch[touchPoint1.id()];
                            speed = QLineF(0,0, 0.196*normalSpeed.x(), 0.147*normalSpeed.y()).length();  //values between 0 and 1, converted to m/s
                        }else{
                            if(times.contains(touchPoint1.id())){
                                QTime *time = times[touchPoint1.id()];
                                int ms = time->elapsed();
                                time->restart();
                                if(ms != 0)
                                    speed = QLineF(touchPoint1.scenePos(), touchPoint1.lastScenePos()).length()*px2mmScreen / (double)ms;  //exprimed in m/s (mm/ms = m/s)
                            }else
                                speed = 0;
                        }
                        speed /= sqrt(GhostView::scaleFactor);

                        float cdgain = 0;  //values between 1 and 7
                        if(speed > 1)
                            cdgain = 6;
                        else if(speed < 0)
                            cdgain = 0;
                        else
                            cdgain = -12*pow(speed, 2) + 17*speed + 1;  //default gain in OS X touchpad

                        QPointF tmp((touchPoint1.scenePos().x() - touchPoint1.lastScenePos().x()) * sizeTouchpad.width()  / (x_width *px2mmScreen),
                                    (touchPoint1.scenePos().y() - touchPoint1.lastScenePos().y()) * sizeTouchpad.height() / (y_height*px2mmScreen));
                        cursors.begin().value()->moveBy(cdgain * tmp.x(),
                                       cdgain * tmp.y());
                        cursors.begin().value()->moveCursel(cursors.begin().value()->getFirstId(), QPoint());

                    }
                }
                break;

            case Qt::TouchPointReleased:
                receiveTouchEnd();
                if(memoryLeftClick == touchPoint1.id()){
                    if(touchIdTOcurselId.isEmpty())
                        cursors.begin().value()->setLeftClick(false);
                    memoryLeftClick = -1;

                }else{
                    //click direct touch
                    if(QLineF(touchPoint1.pos(), touchPoint1.startPos()).length() <= MIN_DP
                         && timerClickList.contains(touchPoint1.id())){
                        //qDebug("generate cursel click");
                        timerClickList[touchPoint1.id()]->deleteLater();
                        timerClickList.remove(touchPoint1.id());
                        clickPosition.remove(touchPoint1.id());

                        int fId = cursors.begin().value()->addCursel((touchPoint1.pos() + QPointF(0, (BUTTON_LINE+BUTTON_HEIGHT)/2.0))/GhostView::scaleFactor);
                        if(fId != -1){
                            if(touchIdTOcurselId.isEmpty())
                                cursors.begin().value()->setLeftClick(true);
                            touchIdTOcurselId.insert(touchPoint1.id(), fId);
                        }
                    }

                    //Remove direct touch
                    if(touchIdTOcurselId.contains(touchPoint1.id())){
                        //qDebug("release cursel");
                        cursors.begin().value()->removeCursel(touchIdTOcurselId[touchPoint1.id()], (touchPoint1.pos() + QPointF(0, (BUTTON_LINE+BUTTON_HEIGHT)/2.0))/GhostView::scaleFactor);
                        touchIdTOcurselId.remove(touchPoint1.id());
                        if(touchIdTOcurselId.isEmpty() && memoryLeftClick == -1){
                            timerClutch = new QTimer(this);
                            timerClutch->setSingleShot(true);
                            connect(timerClutch, SIGNAL(timeout()), this, SLOT(stopCursel()));
                            timerClutch->start(500);
                        }

                    //Control of cursor
                    }else if(controlMouse){
                        //qDebug("end mouse");
                        controlMouse = false;
                        if(times.contains(touchPoint1.id()))
                            delete times[touchPoint1.id()];
                        times.remove(touchPoint1.id());
                    }
                }
                break;

            default:
                break;
            }
        }
        cursors.begin().value()->generateEvent();
        break;

    default:
        return QGraphicsItem::sceneEvent(event);
    }
    update();
    return true;
}

void SoftTouchpad::startCursel(){
    //qDebug("press cursel");
    QTimer *timerClick = (QTimer *)sender();
    int cursorId = timerClickList.key(timerClick);

    setOpacity(1);
    directTouch = true;
    if(timerClutch){
        timerClutch->stop();
        timerClutch->deleteLater();
        timerClutch = NULL;
    }
    int fId = cursors.begin().value()->addCursel((clickPosition[cursorId] + QPointF(0, (BUTTON_LINE+BUTTON_HEIGHT)/2.0))/GhostView::scaleFactor);
    if(fId != -1){
        if(touchIdTOcurselId.isEmpty())
            cursors.begin().value()->setLeftClick(true);
        touchIdTOcurselId.insert(cursorId, fId);
    }

    timerClick->deleteLater();
    timerClickList.remove(cursorId);
    clickPosition.remove(cursorId);

    update();
}

void SoftTouchpad::stopCursel(){
    cursors.begin().value()->setLeftClick(false);
    setOpacity(0.6);
    directTouch = false;
}
